/**
* File         : BitArray.cc 
* Author       : Zeynep Odcikin-Ozdemir
* Created      : Friday Feb 20 09:23:06 1998
* Revision     : $Revision$
* Updated      : $Date$
*-----------------------------
*  BitArray Object Definition.
*-----------------------------
*/

#include <iostream.h>
#include <fstream.h>
#include <math.h>

#include "DESError.h"
#include "BitArray.h"
#include "deffunc.h"

/** 
* Default Constructor for this object 
* Since there is no information, then everything is empty.
*
*
*         size : denotes the number of bits
*
*  numElements : denotes the number of elements we need to keep these
*                bits in the array
*
*         bits : is a pointer to this array.
*
*/
template<class T>
BitArray<T>::BitArray() 
{
   size = 0 ;
   numElements = 0;
   bits = NULL;

}// end of Constructor

/** 
*  Construct a bit array of size sz 
*  It defines 'sz' number of bits.
*
*         size : denotes the number of bits
*
*  numElements : denotes the number of elements we need to keep these
*                bits in the array
*
*         bits : is a pointer to this array.
*
*/
template<class T>
BitArray<T>::BitArray(size_t sz)
{
   size = sz ;
   numElements = 0;
   bits = NULL;

   /*********************************************************************/
   /** If size is greater than zero, we need to define other attributes */
   /*********************************************************************/
   if(size>0)
   {
      numElements = (size-1)/NumBitsPerElement + 1;
      bits = new T[numElements];
      for(size_t i=0;i<numElements;i++) bits[i] = (T)0;

   }// end of if

}// end of Constructor

/** 
* Copy Constructor 
*/ 
template<class T>
BitArray<T>::BitArray(BitArray<T>& RHS)
{
   copyFrom(RHS);

}// end of COPY Constructor

/** 
* Destructor for BitArray Object 
*/
template<class T>
BitArray<T>::~BitArray()
{
   dispose();
}// end of Destructor

/** 
* Assignment Operator 
*
* It assigns one BitArray to another:
*
* For example: 
*     BitArray a(6);
*     BitArray b(7);
*     ......
*     a = b;
*             // a has 7 bits now    
*    
* It does not apply size checks. Left handside object will be 
* the exact copy of the right handside object.
*
*/
template<class T>
BitArray<T> BitArray<T>::operator=(const BitArray& RHS) 
{
   dispose();
   copyFrom(RHS);
   return (*this);

}// end of Assignment Operation

/** 
*
*  Returns the number of elements in the bits array.
*
*  Note: This is NOT the number of bits.
*
*/
template<class T>
size_t BitArray<T>::getNumberOfElements() const
{ 
   return numElements; 
}

/** 
*
*  Returns the number of bits in this bit array.
*
*/
template<class T>
size_t BitArray<T>::getNumberOfBits() const
{ 
   return size; 
}

/** 
*
*  Sets the given bit to 1. If the index(n) is outof range, 
*  then it will throw an exception.
*
*  Inputs:
*      n : index of the bit
*
*  Exceptions:
*      "Index out of bound"
*
*/
template<class T>
void BitArray<T>::set(size_t n)
{ 
   char mess[80];
   sprintf(mess,"Index out of bound. Index is %d ",n);

   if(n>=0 && n<size)  
      bits [n/NumBitsPerElement] |=  ( (T) 1 << n % NumBitsPerElement );
   else throw DESError(__FILE__,__LINE__,mess);

}// end of set()

/** 
*
*  Sets the given bit to 0. If the index(n) is outof range, 
*  then it will throw an exception.
*
*  Inputs:
*      n : index of the bit
*
*  Exceptions:
*      "Index out of bound"
*
*/
template<class T>
void BitArray<T>::reset(size_t n)
{ 
   char mess[80];
   sprintf(mess,"Index out of bound. Index is %d ",n);

   if(n>=0 && n<size) 
      bits [n/NumBitsPerElement] &=  ~( (elttype) 1 << (n%NumBitsPerElement) );
   else throw DESError(__FILE__,__LINE__,mess);

}// end of reset()

/** 
*  Tests whether the indexed bit is 1 or not.
*  If it is value is 1, it will return TRUE(1),
*  Otherwise it will return FALSE(0).
*

*  Inputs:
*      n : index of the bit
*
*  Output:
*      TRUE : if the indexed bit's value is 1
*     FALSE : if the indexed bit's value is 0
*
*  Exceptions:
*      "Index out of bound"
*

*/
template<class T>
Boolean BitArray<T>::test(size_t n) const
{ 
   char mess[80];
   sprintf(mess,"Index out of bound. Index is %d ",n);

   /**********************/
   /** apply range check */
   /**********************/
   if(n<0 && n>size) throw DESError(__FILE__,__LINE__,mess);

   /***********************************************************************/
   /** Find the element, which contains the related bit, in the array     */
   /** Define an element whose only one bit's value is 1                  */ 
   /** apply bitwise AND operation.                                       */
   /** The result will tell you whether corresponding bit is set or not.  */
   /***********************************************************************/
   int z = bits [n/NumBitsPerElement] & ( (T) 1 << (n%NumBitsPerElement) );
   return (z?TRUE:FALSE);

}// end of test()

template<class T>
size_t BitArray<T>::count() const
{ 
   size_t tally=0;
   for(size_t i=0; i<size; i++) {if(test(i)) tally++;}
   return tally;
}

/** 
*  Applies permutation array onto the BitArray.
*
*  Inputs:
*      permutation : permutation information, 
*                    It starts from 1 and left to right.
*
*               sz : size of the permutation array
*
*  Output:
*      Permuted BitArray according to the given permutation information.
*
*  Exceptions:
*      "Index out of bound"
*
*  Forexample:
*     a = 10101010 and permutation = { 2,4,6,8,1,3,5,7 }
*     b = a.suffle(permutation,8);
*            // b = 00001111
*/
template<class T>
BitArray<T> BitArray<T>::shuffle(int *permutation, int sz)
{
  
   BitArray<T> ret(sz);

   for(size_t i=0;i<sz ;i++) if(test(permutation[i]-1)) ret.set(i);
    
   return ret;

}// end of shuffle()

/** 
*
*  Converts the value to BitArray. It tries to find the minimum
*  number of bits to represent this value.
*
*  Inputs:
*      value : an unsigned integer value.
*
*  Output:
*      A BitArray representation of the given value.
*
*  Exceptions:
*      "Index out of bound"
*
*  Forexample:
*        a = b.IntToBitArray(0);
*              // a = 0
*
*        a = b.IntToBitArray(1);
*              // a = 1
*
*        a = b.IntToBitArray(5);
*              // a = 101
*
*        a = b.IntToBitArray(15);
*              // a = 1111
*
*/
template<class T>
BitArray<T> BitArray<T>::IntToBitArray(unsigned int value) const
{
   int i;

   /*************************************************************/
   /** Get the number of bits required to represent this number */
   /*************************************************************/
   int sz = (int)((value>0?(log10(value)/log10(2.0)):0.0)+1);
   unsigned int  rem = value;

   BitArray<T> ret(sz);

   for(i=sz-1;i>=0;i--)
   {
      if(rem % 2) ret.set(i);
      rem = rem / 2;

   }// end of for

   return ret;

}// end of IntToBitArray()

/** 
*  Converts the value to BitArray. It tries to put everything in
*  given number of bits(N), if the minimum number of bits to represent 
*  this number is less than N, it places them to 
*  right marigin(least significant bit places to the rightmost element). 
*  If N is less than, the minimum number of bits to store this info,
*  it returns the first N bit from left to right.
*
*  Inputs:
*      value : an unsigned integer value.
*
*  Output:
*      A BitArray representation of the given value.
*
*  Exceptions:
*      "Index out of bound"
*
*  Forexample:
*        a = b.IntToBitArray(0,4);
*              // a = 0000
*
*        a = b.IntToBitArray(1,4);
*              // a = 0001
*
*        a = b.IntToBitArray(5,2);
*              // a = 10
*
*        a = b.IntToBitArray(15,6);
*              // a = 001111
*
*/
template<class T>
BitArray<T> BitArray<T>::IntToBitArray(unsigned int value, int N) const
{
   BitArray<T> b(N);
   BitArray<T> a = IntToBitArray(value);

   int shift = N - a.getNumberOfBits();
   if(shift<0) shift = 0;
   else if(shift>N) shift = (shift%N);

   a = b | a;
   if(shift) a = a>>shift;

   return a;

}// end of IntToBitArray()

/** 
* Converts hexadecimal values in a string binary and store 
* into a BitArray
*/
template<class T>
BitArray<T> BitArray<T>::HexStringToBitArray(char *buf, int s) const
{
   BitArray<T> b(0);
   
   for(size_t i=0;i<s;i++)
   {
      if(buf[i] >= '0' && buf[i]<='9') 
	 b = b + b.IntToBitArray(buf[i]-'0',4);
      else if(buf[i] >= 'A' && buf[i]<='F')
	 b = b + b.IntToBitArray(buf[i]-'A'+10,4);
      else if(buf[i] >= 'a' && buf[i]<='f')
	 b = b + b.IntToBitArray(buf[i]-'a'+10,4);
      else
       {
	  char mess[80];
	  sprintf(mess,"Format Error! %s",buf);
	  cout << " I will throw the exception !" << endl;
	  throw new DESError(__FILE__,__LINE__,mess);
       }

   }// end of for

   return b;

}// end of HexStringToBitArray()

/** 
* BitArray to integer conversion
*/
template<class T>
unsigned int BitArray<T>::toInt() const
{
   int i = size-1;
   unsigned int ret = (test(i)?1:0);
   unsigned int pw  = 2;
   i--;
   for(;i>=0;i--,pw*=2) ret += (test(i)?pw:0);

   return ret;

}// end of toInt()

/** 
*/
template<class T>
BitArray<T> BitArray<T>::copyFrom(size_t from, size_t numberOfBits)
{
  
   BitArray<T> ret(numberOfBits);

   for(size_t i=0;i<numberOfBits && (i+from)<size;i++) 
	 if(test(i+from)) ret.set(i);
    
   return ret;

}// end of copyFrom()

/** 
*/
template<class T>
BitArray<T> BitArray<T>::cyclicCopyFrom(size_t from, size_t numberOfBits)
{
  
   BitArray<T> ret(numberOfBits);
   size_t j=from;

   for(size_t i=0;i<numberOfBits; i++) 
   {
     j = j % size;
     if(test(j)) ret.set(i);
     j++;
     if(j>=size) j=j%size;

   }

   return ret;

}// end of cyclicCopyFrom()

/** 
* And  Operator 
*/
template<class T>
BitArray<T> BitArray<T>::operator&(const BitArray<T>& RHS)
{
  
   //BitArray ret = (*this);
   BitArray<T> ret(maximum(getNumberOfBits(),RHS.getNumberOfBits() ));

   size_t count = minimum ( getNumberOfElements(), RHS.getNumberOfElements() );

   for(size_t i=0;i<count;i++) ret.bits[i] = bits[i] & RHS.bits[i];
    
   return ret;

}// end of operator&()

/** 
* Or  Operator 
*/
template<class T>
BitArray<T> BitArray<T>::operator|(const BitArray<T>& RHS)
{
   //BitArray ret = (*this);
   BitArray<T> ret(maximum(getNumberOfBits(),RHS.getNumberOfBits() ));

   size_t count = minimum ( getNumberOfElements(), RHS.getNumberOfElements() );

   for(size_t i=0;i<count;i++) ret.bits[i] = bits[i] | RHS.bits[i];

   return ret;

}// end of operator|()

/** 
* Exlusive-OR (XOR)  Operator 
*/
template<class T>
BitArray<T> BitArray<T>::operator^(const BitArray<T>& RHS)
{
   //BitArray ret = (*this);
   BitArray<T> ret(maximum(getNumberOfBits(),RHS.getNumberOfBits() ));

   size_t count = minimum ( getNumberOfElements(), RHS.getNumberOfElements() );

   for(size_t i=0;i<count;i++) ret.bits[i] = bits[i] ^ RHS.bits[i];

   return ret;

}// end of operator^()

/** 
* Shift-Left operator
*/
template<class T>
BitArray<T> BitArray<T>::operator<<(size_t n)
{
   return this->copyFrom(n,size);

}// end of operator<<()

/** 
* Shift-Right operator
*/
template<class T>
BitArray<T> BitArray<T>::operator>>(size_t n)
{
   BitArray<T> ret(size);

   for(size_t i=0;i<(size-n) ;i++) if(test(i)) ret.set(i+n);

   return ret;

}//  end of operator>>()

/** 
* @(append) operation on BitArray
*/
template<class T>
BitArray<T> BitArray<T>::operator+(const BitArray<T>& RHS)
{
   BitArray<T> ret(size+RHS.size);

   size_t i ;
   size_t j = 0;

   for(i=0;i<size ;i++,j++) if(test(i)) ret.set(j);

   for(i=0;i<RHS.size ;i++,j++) if(RHS.test(i)) ret.set(j);

   return ret;

}// end of operator+()

/** 
* Overwrite the stream input operation 
*/
template<class T>
istream& operator>>(istream& is,  BitArray<T>& RHS)
{
   char ch;

   for(size_t i=RHS.size;i>0;) 
   { 
      i--;
      is >> ch;

      if(ch == '1' ) RHS.set(i);
      else RHS.reset(i);

   }// end of for

   return is;

}// end of operator>>()

/** 
* Reclaims the memory allocated before for this object 
*/
template<class T>
void BitArray<T>::dispose()
{
   if(numElements>0) delete[] bits;

}// end of dispose()

/** 
* Copies the object to this object 
*/
template<class T>
void BitArray<T>::copyFrom(const BitArray<T>& RHS)
{
   size = RHS.size;
   numElements = RHS.numElements;
   bits = new T[numElements];
   for(size_t i=0;i<numElements;i++) bits[i] = RHS.bits[i];

}// end of copyFrom()

/** 
* Prints the object to the osutput stream 
*/
template<class T>
void BitArray<T>::printMe(ostream& os) const 
{
  printMeClean(os);
  os << " sz : " << size << " "; 

}// end of printMe()

/** 
* Prints the object to the osutput stream 
*/
template<class T>
void BitArray<T>::printMeClean(ostream& os) const 
{
  //printBinary(os);

  BitArray<T> a;
  a = (*this);

  BitArray<T> b;
  int cnt = (size-1)/4 + 1;
  for(int i=0;i<cnt;i++) 
  { 
     b = a.copyFrom((size_t)i*4,(size_t)4);
     os << toHex[b.toInt()];
     //printBinary(os);

  }// end of for()

}// end of printMeClean()

/** 
* Prints the object to the osutput stream 
*/
template<class T>
void BitArray<T>::printBinary(ostream& os) const 
{
  for(size_t i=0;i<size;i++) os << (test(i)?'1':'0');

}// end of printBinary()

